#include <SoftwareSerial.h> 
#include <Servo.h>
#include <FastLED.h>

// PIN numbers
#define BT_SERIAL_RX      2 
#define BT_SERIAL_TX      3 
#define RGB_STRIP_TOP     4 
#define RGB_STRIP_LEFT    5 
#define RGB_STRIP_RIGHT   6 
#define LASER             9  
#define SERVO_1           A0 
#define SERVO_2           A1 
#define SERVO_3           A2 
#define SERVO_4           A3 
#define DISCO_SERVO       A5 

// LED strip constants
#define NUM_LEDS_TOP    6
#define NUM_LEDS_LEFT   9
#define NUM_LEDS_RIGHT  7
#define LED_TYPE        WS2811
#define COLOR_ORDER     GRB

// Bluetooth variables
SoftwareSerial btSerial(BT_SERIAL_RX,BT_SERIAL_TX);
int index = 0; 
char data[12]; 
boolean flag = false;

// Servo variables
boolean isServoOn = false;
Servo servo1;
Servo servo2;
Servo servo3;
Servo servo4;
Servo discoServo;

int stillSpeed = 1500;
int currentServoSpeed1 = stillSpeed;
int currentServoSpeed2 = stillSpeed;
int currentServoSpeed3 = stillSpeed;
int currentServoSpeed4 = stillSpeed;
int currentDiscoServoAngle = 90;
boolean isTurningClockwise = true;

// Light variables
boolean isLedOn = true;
CRGB ledsTop[NUM_LEDS_TOP];
CRGB ledsLeft[NUM_LEDS_LEFT];
CRGB ledsRight[NUM_LEDS_RIGHT];
CRGB currentColor;
boolean isDiscoOn = false;
boolean ledStateBeforeDisco = false;

// Timer variables
long previousMillis1 = 0;
long previousMillis2 = 0;
long previousMillisDiscoServo = 0;
long currentMillis = 0;
long discoLightDelay = 300;
long discoWheelsDelay = 4000;
long discoServoDelay = 10;
boolean isMovingClockwise = true;

void setup() { 
  pinMode(LASER, OUTPUT); 
  digitalWrite(LASER, HIGH);

  // LED setup
  FastLED.addLeds<LED_TYPE, RGB_STRIP_TOP, COLOR_ORDER>(ledsTop, NUM_LEDS_TOP).setCorrection(TypicalLEDStrip);
  FastLED.addLeds<LED_TYPE, RGB_STRIP_LEFT, COLOR_ORDER>(ledsLeft, NUM_LEDS_LEFT).setCorrection(TypicalLEDStrip);
  FastLED.addLeds<LED_TYPE, RGB_STRIP_RIGHT, COLOR_ORDER>(ledsRight, NUM_LEDS_RIGHT).setCorrection(TypicalLEDStrip);
  currentColor = CRGB(0, 0, 0);

  // Servo setup
  discoServo.attach(DISCO_SERVO);
  discoServo.write(currentDiscoServoAngle);
  
  Serial.begin(9600);
  btSerial.begin(9600);
} 

void loop() { 
  char incomingChar; 
  
  if (btSerial.available() > 0) { 
    while (btSerial.available() > 0) { 
      incomingChar = btSerial.read(); 
      delay(10); //Delay required 
      data[index] = incomingChar; 
      index++; 
    }
    data[index] = '\0'; 
    flag = true;   
  } 
  
  if (flag) { 
    processCommand(); 
    flag = false; 
    index = 0; 
    data[0] = '\0'; 
  } 
  
  currentMillis = millis();

  if (isDiscoOn) {
    handleDiscoMode();
  }
  
  // Spin the servo's
  servo1.writeMicroseconds(currentServoSpeed1);
  servo2.writeMicroseconds(currentServoSpeed2);
  servo3.writeMicroseconds(currentServoSpeed3);
  servo4.writeMicroseconds(currentServoSpeed4);

  if (currentDiscoServoAngle <= 0 || currentDiscoServoAngle >= 180) {
    isTurningClockwise = !isTurningClockwise;
  }
  
  if (currentMillis - previousMillisDiscoServo >= discoServoDelay) {
    previousMillisDiscoServo = currentMillis;
    
    if (isTurningClockwise) {
      currentDiscoServoAngle++;
    } else {
      currentDiscoServoAngle--;
    }
  }
  
  discoServo.write(currentDiscoServoAngle);

  delay(1);
} 

// Processes commands received from the bluetooth connection with the Android app
void processCommand() { 
  Serial.println(data);
  
  char command = data[0]; 
  int index = 1;
  char dataAtIndex = data[index];
  
  // Handles LED light commands 
  if (command == 'L' && !isDiscoOn) { 
    if (dataAtIndex == '1') { 
      setLedPower(true);
    } 
    else if (dataAtIndex == '0') { 
      setLedPower(false);
    }
  } 
  
  // Handles laser commands
  else if (command == 'B' && !isDiscoOn) { 
    if (dataAtIndex == '1') { 
      digitalWrite(LASER, HIGH); 
    } 
    else if (dataAtIndex == '0') { 
      digitalWrite(LASER, LOW); 
    }
  } 
  
  // Handles changing the color of LED lights
  else if (command == 'C' && !isDiscoOn) { 
    String red = "";
    String green = "";
    String blue = "";
    boolean isRedRead = false;
    boolean isGreenRead = false;
    
    while (index <= 12) {
      dataAtIndex = data[index];
      
      if (dataAtIndex == '|') {
        if (!isRedRead) {
          isRedRead = true;
        } else {
          isGreenRead = true;
        }
      } else {
        if (!isRedRead && !isGreenRead) {
          red += dataAtIndex;
        } else if (isRedRead && !isGreenRead) {
          green += dataAtIndex;
        } else {
          blue += dataAtIndex;
        }
      }
  
      index++;
    }
  
    currentColor = CRGB(red.toInt(), green.toInt(), blue.toInt());
    updateLedStrip(currentColor);
  } 
  
  // Handles disco commands
  else if (command == 'D') {
    if (dataAtIndex == '1') { 
      startServoMotors();
      isDiscoOn = true;
      ledStateBeforeDisco = isLedOn;
      previousMillis1 = millis();
      previousMillis2 = millis();
    } 
    else if (dataAtIndex == '0') { 
      stopServoMotors();
      isDiscoOn = false;
      isLedOn = ledStateBeforeDisco;
      setLedPower(isLedOn);
    } 
  } else if (command == 'J' && !isDiscoOn) {
    if (!isServoOn) {
      startServoMotors(); 
    }
    
    String joystickAngle = "";
    String joystickStrength = "";
    boolean isAngleRead = false;
    
    while (index <= 12) {
      dataAtIndex = data[index];
      
      if (dataAtIndex == '|') {
        isAngleRead = true;
      } else {
        if (isAngleRead) {
          joystickStrength += dataAtIndex;
        } else {
          joystickAngle += dataAtIndex;
        }
      }
    
      index++;
    }

    if (joystickStrength.toInt() > 0) {
      updateServoMotors(joystickAngle.toInt(), joystickStrength.toInt());
    } else {
      stopServoMotors();
    }
  }
} 

// Toggles the LED power between on and off
void setLedPower(boolean ledPower) {
  if (ledPower) { 
      updateLedStrip(currentColor);
    } 
    else { 
      CRGB black = CRGB(0, 0, 0);
      updateLedStrip(black);
    }
}

// Updates the LED strip according to data from the colorpicker
void updateLedStrip(CRGB color) {

  // Change color of top LED strip
  for (int i = 0; i <= NUM_LEDS_TOP; i++) {
    ledsTop[i] = color;
  }
  
  // Change color of left LED strip
  for (int i = 0; i <= NUM_LEDS_LEFT; i++) {
    ledsLeft[i] = color;
  }

  // Change color of right LED strip
  for (int i = 0; i <= NUM_LEDS_RIGHT; i++) {
    ledsRight[i] = color;
  }

  FastLED.show();
}

// Updates the speed of every servo motor according to data from the joystick
void updateServoMotors(int angle, int strength) {
  strength = 5*strength;

  if (70 <= angle && angle <= 110) {
    currentServoSpeed1 = stillSpeed - strength;
    currentServoSpeed3 = stillSpeed + strength;
    currentServoSpeed2 = stillSpeed;
    currentServoSpeed4 = stillSpeed;
  } else if (250 <= angle && angle <= 290) {
    currentServoSpeed1 = stillSpeed + strength;
    currentServoSpeed3 = stillSpeed - strength;
    currentServoSpeed2 = stillSpeed;
    currentServoSpeed4 = stillSpeed;
  } else if (160 <= angle && angle <= 200) {
    currentServoSpeed1 = stillSpeed;
    currentServoSpeed3 = stillSpeed;
    currentServoSpeed2 = stillSpeed - strength;
    currentServoSpeed4 = stillSpeed + strength;
  } else if (340 <= angle || angle <= 20) {
    currentServoSpeed1 = stillSpeed;
    currentServoSpeed3 = stillSpeed;
    currentServoSpeed2 = stillSpeed + strength;
    currentServoSpeed4 = stillSpeed - strength;
  }
}

// Handles disco mode
void handleDiscoMode() {  
  if (currentMillis - previousMillis1 >= discoLightDelay) {
    previousMillis1 = currentMillis;
    int red = random(0, 255);
    int green = random(0, 255);
    int blue = random(0, 255);
    CRGB randomColor = CRGB(red, green, blue);
    updateLedStrip(randomColor);
  }

  if (currentMillis - previousMillis2 >= discoWheelsDelay) {
    previousMillis2 = currentMillis;

    int turningSpeed = random(400, 500);
    
    if (isMovingClockwise) {
      turningSpeed = -turningSpeed;
    }
    isMovingClockwise = !isMovingClockwise;
    
    currentServoSpeed1 = stillSpeed + turningSpeed;
    currentServoSpeed2 = stillSpeed + turningSpeed;
    currentServoSpeed3 = stillSpeed + turningSpeed;
    currentServoSpeed4 = stillSpeed + turningSpeed;
  }
}

// Starts servo motors
void startServoMotors() {
  isServoOn = true;
  servo1.attach(SERVO_1);
  servo2.attach(SERVO_2);
  servo3.attach(SERVO_3);
  servo4.attach(SERVO_4);
  discoServo.detach();
}

// Stops servo motors
void stopServoMotors() {
  isServoOn = false;
  currentServoSpeed1 = stillSpeed;
  currentServoSpeed2 = stillSpeed;
  currentServoSpeed3 = stillSpeed;
  currentServoSpeed4 = stillSpeed;
  servo1.detach();
  servo2.detach();
  servo3.detach();
  servo4.detach();
  discoServo.attach(DISCO_SERVO);
}
